#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Forger has two tasks: faking CASA logs and faking CASA products
"""
import json
import logging
import os
import random
import tarfile
from pathlib import Path

import arrow
import vela.content.forger_content as content


class VelaLog:
    def __init__(self, metadata: str, product_type: str):
        self.logger = logging.getLogger("vela")
        self.metadata = metadata
        self.product_type = product_type

    def forge_logs(self, parameters: dict):
        # forging three files. 2 for pipeline completion, 1 log file
        # vela-pipeline.sh.out.txt
        # vela-pipeline.sh.err.txt
        # vela-<date_as_string>-<timestring>.log

        workflow_name = parameters["workflowName"]

        if "imaging" in workflow_name:
            self.img_logs()
        elif "calibration" or "restore" in workflow_name:
            self.cal_logs()

    def cal_logs(self):
        """
        forge calibration type logs

        :return:
        """
        path = Path.cwd() / "working"
        os.chdir(path)

        # make log files
        pipe_out = open("vela-pipeline.sh.out.txt", "x")
        pipe_out.write(forged_content())
        pipe_out.close()

        pipe_err = open("vela-pipeline.sh.err.txt", "x")
        pipe_err.write(forged_content())
        pipe_err.close()

        datestring = arrow.now().format("YYYYMMDD")
        timestring = arrow.utcnow().format("hhmmss")

        log = open(f"vela-{datestring}-{timestring}.log", "w")
        log.write(forged_content(self.product_type))
        log.close()

        self.logger.info("Logs forging complete.")
        os.chdir("../")

    def img_logs(self):
        """
        forge imaging type logs

        :return:
        """
        # TODO: find out if logs are different for imaging
        self.cal_logs()


class VelaProduct:
    def __init__(self, metadata: str, product_type: str):
        self.logger = logging.getLogger("vela")
        self.metadata = metadata
        self.product_type = product_type

    def forge_products(self, parameters: dict):
        self.logger.info(f"product type: {self.product_type}")
        if self.product_type in ["vlass_img", "standard_img", "integrated"]:
            self.img_products(parameters)
        elif self.product_type in ["vlass_cal", "standard_cal", "restore"]:
            self.cal_products(parameters)

    def cal_products(self, parameters: dict):
        # forge products so there is something to test delivery with

        path = Path.cwd() / "products"
        os.chdir(path)

        sdm_id = parameters["fileSetIds"][0] if isinstance(parameters["fileSetIds"], list) else parameters["fileSetIds"]
        filename = sdm_id + ".ms.calapply.txt"

        if not Path("./" + filename).exists():
            calapply = open(filename, "x")
            calapply.write("I am a calibration file.")
            calapply.close()

        os.chdir("../working")
        self.forge_measurement_set(sdm_id)
        self.forge_weblog(os.getcwd())

        self.logger.info("Forging calibration products complete.")
        os.chdir("../")

    def img_products(self, parameters: dict):
        """
        Forge a variety of image files for ingestion testing

        :param parameters:
        :return:
        """
        path = Path.cwd() / "products"
        os.chdir(path)

        # pbcor.tt0 is the primary science product
        pbcor_filename = "oussid.1-93305_sci.L_band.cont.I.pbcor.tt0.fits"
        selfcal_pbcor = "oussid.1-93305_sci.S_band.cont.selfcal.I.pbcor.tt0.fits"
        pb_filename = "oussid.1-93305_sci.L_band.cont.I.pb.tt0.fits"
        cont_filename = "oussid.1-93305_sci.L_band.cont.I.tt0.fits"
        mask_filename = "oussid.1-93305_sci.L_band.cont.I.mask.fits"

        fits = open(pbcor_filename, "x")
        fits.write("I am a primary science image.")
        fits.close()

        fits = open(selfcal_pbcor, "x")
        fits.write("I am a selfcal image.")
        fits.close()

        fits = open(pb_filename, "x")
        fits.write("I am a primary beam image.")
        fits.close()

        fits = open(cont_filename, "x")
        fits.write("I am a continuum image.")
        fits.close()

        fits = open(mask_filename, "x")
        fits.write("I am a mask image.")
        fits.close()

        os.chdir("../working")
        self.forge_weblog(os.getcwd())

        self.logger.info("Forging image products complete.")
        os.chdir("../")

    def forge_measurement_set(self, sdm_id: str):
        self.logger.info("Forging Measurement Set....")
        cms_name = sdm_id + ".ms"

        if not Path("./" + cms_name).exists():
            os.mkdir(cms_name)
            os.chdir("./" + cms_name)

            for i in range(0, 5):
                dirname = f"FORGED_MS_DIR_{i}"
                os.mkdir(dirname)
                dir_file = open(f"./{dirname}/table.f{i}", "x")
                dir_file.write("I am a file.")
                dir_file.close()
                file = open(f"table.f{i}", "x")
                file.write("I am a table.")
                file.close()

            os.chdir("../")
            self.logger.info("MS forging complete.")
        else:
            os.chdir("../")
            self.logger.info("MS already present, no need to forge")

    def forge_weblog(self, path):
        self.logger.info("Forging weblog...")

        datestring = arrow.now().format("YYYYMMDD")
        timestring = arrow.utcnow().format("hhmmss")
        dirname = f"pipeline-{datestring}-{timestring}"
        os.mkdir(dirname)
        os.chdir(path + "/" + dirname)
        os.mkdir("html")
        os.chdir(path + "/" + dirname + "/html")

        file1 = open(f"vela-{datestring}-{timestring}.log", "x")
        file1.write("I am a weblog file")
        file1.close()

        file2 = open("index.html", "x")
        file2.write("I am a weblog file")
        file2.close()

        os.chdir("../../../products")

        with tarfile.open("weblog.tgz", "w:gz") as tar:
            tar.add(path + "/" + dirname, arcname=os.path.basename(dirname))
            tar.close()

        self.logger.info("Weblog forging complete.")


def read_metadata(filename: str) -> dict:
    with open(filename) as file:
        return json.loads(file.read())


def forged_content(product_type: str = None) -> str:
    forged = content.file_content()

    if product_type and product_type == "restore":
        # Decide of we're going to randomly fail for testing
        fail = random.choice([True, False])
        print(f"forging failure? {fail}")
        forged = (forged + content.restore_failure_log_content()) if fail else forged

    return forged


# write fake logs and products
def forge(metadata: str, product_type: str):
    parameters = read_metadata(metadata)
    fake_logs = VelaLog(metadata, product_type)
    fake_logs.forge_logs(parameters)

    fake_products = VelaProduct(metadata, product_type)
    fake_products.forge_products(parameters)
